       integer function ffunct()
c       Program MidPtFM2d (wrap)
c
c   fractal generator adapted from algorithm MidPointFM2D in the 
c     Science of Fractal Images, M.F. Barnsley, R. L. Devaney
c     B. B. Mandelbrot, H.-O. Peitgen, D. Saupe, and R. F. Voss
c     Springer-Verlag
c
c  this program modified to combine multiple fractal topographies
c     then normalize them relative to each other,
c     and perform multiple horizontal slices
c     to obtain requested p values for each habitat
c     in a new synthetic realization of the landscape
c
c     William W. Hargrove
c     Paul M Schwartz
c     5/7/96
c     (423) 241-2748
c
       implicit real*4 (a-h,o-z)
       parameter (maxN=515)
       parameter (maxsp=10)
       external sleep
       external time
       integer time
       integer*2 P(maxN*maxN,0:maxsp+3)
       real*4 X(0:maxN,0:maxN)
       real*4 prob(0:maxsp)
       integer*2 hslice(0:maxsp)
       real*4 elev(maxN*maxN), H(maxN,maxsp)
       real*4 northeast, southeast, southwest, northwest
       integer*2 II(0:maxN,0:maxN)
       integer*2 kcnt(0:maxsp),ties(maxsp)
       integer*2 ifcnt,ihabs,iend,istart
       real*4 sigma
       integer*2 maxlevel, N
       integer*2 maxgray
       integer*4 iseed
       integer*2 idispvec(0:maxN*maxN)
       character*1 ans
       character*28 maptitle
       character*60 mapfile
       character*10 mapcolor
       logical*1 addition, wrap, gradient, slice, invert
       logical*1 grassout, xpmout, probout, vizout
       logical*1 constraint(0:maxsp)
c
       common/blkihab / ihabs,numcells,itiedsplit,
     &  irowcols,mtsplit,mtresolved,isupply,maxgray
c
       common/blksd/ iseed
       common/blkcrnrs/ northeast, southeast, southwest, northwest
c
c  write execution script for next run
       open (9, file='input.scr',status='unknown')
c
c
c enter random number seed
       type 196
196    format (t10,'Do you want to enter a random number seed <Y>',
     &  /t12,' or use the system clock? [clock]')
       accept 97, ans
97     format (a1)
       if (ans .eq. 'Y' .or. ans .eq. 'y') then
          write (9,197) ans
197       format (a1,t20,"(enter seed or clock?)")
          type 100
100       format (t10,'Enter random number seed ')
          accept *, iseed
          write (9,101) iseed
101       format (i12,t20,"(random number seed)")
       else
21        iseed = time( )
          if (iseed .eq. 0) then
             print *, "iseed was zero."
             go to 21
          endif
          type 298
298       format (t15,'Random number seed from system clock.'/)
          write (9,29) ans, iseed
29        format (a1,t20,"(random seed from clock = ",i12,")")
       endif
c
       type 1017, iseed
1017   format (t15,'Random number seed = ',i12/)
c
c
       type 512
512    format (t10,'Do you want to visualize',
     & /t12, ' the synthetic landscape? <Y,N>')
       accept 522, ans
522    format (a1)
       write (9,57) ans
57     format (a1,t20,"(visualize the landscape?)")
       if (ans. eq. 'y' .or. ans .eq. 'Y') then
         vizout = .true.
       else
         vizout = .false.
       endif
c
c
1      type 225
225    format (t10,'How many levels? (map is 2**levels on a side)')
       accept *, maxlevel
c
       N = 2**maxlevel
       irowcols = N
       numcells = (N+1)*(N+1)
       if (N .gt. maxN) goto 1
       write (9,102) maxlevel
102    format (i3,t20,"(Number of levels)")
c
       type 114
114    format (t10,'Highest data category (excluding no-data)',
     & /t12,' in the realized synthetic landscape?')
       accept *, ihabs
       write (9,199) ihabs
199    format (i3,t20,"(highest data category)")
c
       type 232
232    format (t10,'Will you supply external probability maps',
     & /t12,' for any of these categories?')
       accept 111, ans
111    format (a1)
       write (9,161) ans
161    format (a1,t20,"(supply external prob maps?)")
c
       do j = 0, ihabs
         constraint(j) = .false.
       enddo
c
       isupply = 0
       if (ans .eq. 'Y' .or. ans .eq. 'y') then
c
         type 134
134      format (t10,'For how many categories in the synthetic',
     &   ' landscape',/t12,' will you supply external probability',
     &   '  maps?')
         accept *, isupply
         write (9,189) isupply
189      format (i3,t20,"(number of external prob maps)")
c
         do i = 1, isupply
           type 216
216        format (t10,'Constrain which map category?')
           accept *, k
           write (9,799) k
799        format (i3,t20,"(constrain which category?)")
c          set constraint flag for this category
           constraint(k) = .true.
c
           type 220
220        format (t10,'Enter name of input map file (integer*2, ',
     &     /t12,' delimited by single spaces): ')
           accept 230, mapfile
230        format (a60)
           type 2030, mapfile
2030       format (t15,'Input map name: ',a60/)
           write (9,231) mapfile
231        format (a60)
c
           type 144
144        format (t10, 'Invert probabilities (i.e., DEM)? <Y,N>')
           accept 321, ans
321        format (a1)
           write (9,327) ans
327        format (a1,t20,"(invert probabilities?)")
           if (ans. eq. 'y' .or. ans .eq. 'Y') then
             invert = .true.
           else
             invert = .false.
           endif
c
c          load constraint maps in proper column of P array
c          invert probabilities if necessary, i.e., a DEM
           Call InputMap(mapfile,P,N,k,invert)
         enddo
c
       endif
c
c
       do iz = 0, ihabs
         if (constraint(iz)) goto 664
         do i = 1, maxlevel
118        type 159, i, iz
159        format (t10,'Input value of H for level ',i3,
     &      ', category ',i3)
              accept *, H(i,iz)
              type *, H(i,iz)
              write (9,155) H(i,iz), i, iz
155           format (f4.2,t20,"(value of H for level ",
     &         i2,", category", i3,")")
              if (H(i,iz) .lt. 0.0 .or. H(i,iz) .gt. 1.0) goto 118
         enddo
664    enddo
c
c
       ssum = 0.0
       do i = 1, ihabs
115      type 116, i
116      format (t10,'Probability for habitat ',i3)
         accept *, prob(i)
         if (prob(i) .lt. 0.0) goto 115
         if (prob(i) .gt. 1.0) goto 115
         ssum = ssum + prob(i)
         if (ssum .gt. 1.0000) then
           print *, 'Total probability exceeds 1.0 - Re-enter'
           ssum = ssum - prob(i)
           goto 115
         endif
         write (9,151) prob(i), i
151      format (f6.5,t20,"(probability of habitat type", i2,")")
       enddo
c
c      find greatest non-constraint p to make background
       ztemp = 0.00000
       igreatest = 0
       do i = 0, ihabs
         if (constraint(i)) goto 662
         if (prob(i) .gt. ztemp) then
           ztemp = prob(i)
           igreatest = i
c           print *, 'at i of ',i ,'new greatest is ', igreatest
         endif
662    enddo
c
c      zero category by difference
       prob(0) = 1.0 - ssum
       if (prob(0) .lt. 0.00001) prob(0) = 0.0
       print *, 'prob(0) is  ', prob(0)
c
c       print *, 'igreatest is ', igreatest
       if (prob(0) .gt. prob(igreatest)) igreatest = 0
       print *, 'Cat ', igreatest,' will be calculated by absence'
c
c      The Fractal Realizer assumes that the user-requested category
c with the greatest coverage on the landscape (the highest requested p)
c is the background or matrix category.  In order to give the other
c categories the greatest freedom for fractal tuning and to minimize
c contention ties, this majority background color is computed by absence,
c and all other categories are reclassed.  Furthermore, categories which
c are requested but for which the requested p = 0.0 are not actually
c computed.
c
c      calculate reclassed categories
c       ihabs = 0
c       do i = 0, ihabsmax
c         if (i .eq. igreatest) then
c           ireclass(0) = i  
c           print *, 'Cat ', i,' reassigned to background'
c           goto 215
c         endif
c         if (prob(i) .eq. 0.0) then
c           print *, 'Cat ', i,' requested but not used - skipping'
c           goto 215
c         endif
c         ihabs = ihabs + 1
c         ireclass(ihabs) = i
c         print *, 'cat ', i,' will now be cat ', ihabs
c
c215    enddo
c
c
c       do i = 0, ihabs
c         type 2002, ireclass(i), i, prob(ireclass(i))
c2002     format ('Original category ', i2, ' re-mapped to ',
c     &    i2, ' with requested probability of ', f6.4)
c       enddo
c
c
       type 212
212    format (t10,'Generate GRASS r.in.ascii file? <Y,N>')
       accept 222, ans
222    format (a1)
       write (9,27) ans
27     format (a1,t20,"(generate GRASS input file?)")
       if (ans. eq. 'y' .or. ans .eq. 'Y') then
         grassout = .true.
       else
         grassout = .false.
       endif
c
       type 122
122    format (t10,'Generate xpm file of synthetic landscape? <Y,N>')
       accept 123, ans
123    format (a1)
       write (9,124) ans
124    format (a1,t20,"(xpm file of realized landscape?)")
       if (ans. eq. 'y' .or. ans .eq. 'Y') then
         xpmout = .true.
       else
         xpmout = .false.
       endif
c
       type 322
322    format (t10,'Generate xpm file of fractal probability'
     & ' landscapes? <Y,N>')
       accept 323, ans
323    format (a1)
       write (9,324) ans
324    format (a1,t20,"(xpm file of probability landscapes?)")
       if (ans. eq. 'y' .or. ans .eq. 'Y') then
         probout = .true.
       else
         probout = .false.
       endif
c
c *****************************************************************
c  fill P array with a separate fractal probability map
c    for each desired habitat in the landscape
       do iz = 0, ihabs
         if (constraint(iz)) goto 22
c
c        don't bother if igreatest or no cells requested of this category
         if (prob(iz) .eq. 0.0 .or. iz .eq. igreatest) then
           do i = 1, numcells
             P(i,iz) = 0
           enddo
           goto 22
         endif
c
c       type 102
c102    format (t10,'Input the value for sigma')
c       accept *, sigma
       sigma = 1.
       type 103
103    format (t10,'Input the threshold for detection (xlow) ')
       accept 119, xlow
119    format (g12.6)
       write (9,125) xlow
125    format (f12.6,t20,"(threshold for detection (xlow))")
       if (xlow .eq. 0) xlow = -1e30
       print *, xlow
       type 104
104    format (t10,'Wrap? <Y,N>')
       accept 145, ans
145    format (a1)
       write (9,198) ans
198    format (a1,t20,"(wrap?)")
       if (ans. eq. 'y' .or. ans .eq. 'Y') then
         wrap = .true.
       else
         wrap = .false.
       endif
c
c  no gradient possible with wrap
       if (.not. wrap) then
         type 106
106      format (t10,'Gradient? <Y,N>')
         accept 107, ans
107      format (a1)
       write (9,148) ans
148    format (a1,t20,"(gradient?)")
         if (ans. eq. 'y' .or. ans .eq. 'Y') then
               gradient = .true.
c
           type 1850
1850       format (t10,'Gradient: Initial values for four map corners')
           type 1851
1851       format (t10,'Northeast? ')
           accept *, northeast
           type 1852, northeast
1852       format (5x,g12.6)
           write (9,135) northeast
135        format (f12.6,t20,"(Gradient: northeast)")
c
           type 1853
1853       format (t10,'Southeast? ')
           accept *, southeast
           type 1852, southeast
           write (9,136) southeast
136        format (f12.6,t20,"(Gradient: southeast)")
c
           type 1854
1854       format (t10,'Southwest? ')
           accept *, southwest
           type 1852, southwest
           write (9,137) southwest
137        format (f12.6,t20,"(Gradient: southwest)")
c
           type 1855
1855       format (t10,'Northwest? ')
           accept *, northwest
           type 1852, northwest
           write (9,138) northwest
138        format (f12.6,t20,"(Gradient: northwest)")
c
         else gradient = .false.
         endif
c
       endif	!wrap = .false.
c
c       type 111
c111    format (t10,'Slice into discrete habitats? <Y,N>')
c       accept 110, ans
c110    format (a1)
c       if (ans. eq. 'y' .or. ans .eq. 'Y') then
c             slice = .true.
c         else slice = .false.
c       endif
c
       slice = .false.
c
c
c       type 104
c 104    format (t10,'Addition <Y,N>?')
c        accept 105, ans
c 105    format (a1)
c        if (ans. eq. 'y' .or. ans .eq. 'Y') then 
c              addition = .true.
c          else addition = .false.
c        endif
       addition = .false.
c
c
c  call 2d multi-fractal generator
       print *,"Calling fractal generator for category ", iz
       call mpfm2d (X, maxlevel, sigma, H, addition, iseed, wrap,
     &  gradient,iz)
c
       print *, "fractal generated"
       xmin = 1e30
       xmax = -1e30
       xmean = 0.0
       xss = 0.0
       icnt = 0
c
c  obtain xmin and xmax for range normalization
c   calculate initial raw statistics
c   while you're at it
       do i = 0,N
         do j = 0,N
           if (X(i,j) .lt. xmin) xmin = X(i,j)
           if (X(i,j) .gt. xmax) xmax = X(i,j)
c           xmean = xmean + X(i,j)
c           xss = xss + X(i,j)*X(i,j)
c           icnt = icnt+1
         enddo
       enddo
c       xmean = xmean / icnt
c       type 200, xmean, xss, xmin, xmax
c200    format (/t10,'Hfract Summary'/
c     &          t10,'Mean =    ',g12.6
c     &         /t10,'Sum Sq. = ',g12.6
c     &         /t10,'Minimum = ',g12.6
c     &         /t10,'Maximum = ',g12.6)
c
c  rescale and normalize fractal map between 0 and 32767, inclusive
c
       range = xmax - xmin
       do i = 0,N
         do j = 0,N
           iv = (i*(N+1))+j+1
           P(iv,iz) = ((X(i,j) - xmin)/range) * 32767
           P(iv,ihabs+1) = i
           P(iv,ihabs+2) = j
c          write(10,*) P(iv,iz)
         enddo
       enddo
       print*,'done normalizing now'
c
cc      write GRASS ascii map
c       write(10,791), N
c791    format ('north: ',i4)
c       write(10,792)
c792    format ('south: 0')
c       write(10,793), N
c793    format ('east: ',i4)
c       write(10,794)
c794    format ('west: 0')
c       write(10,795), N+1
c795    format ('rows: ',i4)
c       write(10,796), N+1
c796    format ('cols: ',i4)
c       do i = 0,N
c           write(10,*) (P((i*(N+1))+j+1,iz), j=0,N)
c       enddo
c
c
c  end of iz loop over all landscape categories in synthetic map
22      enddo
c **********************************************************************
c
c  ihabs fractal maps now installed in P matrix:
c
c                 ihabs cols + 3
c		|---------------|
c		|		|
c		|		|
c		|  empty cells  |
c  mtresolved =>|		|
c		|		|
c   mtsplit   =>|---------------|
c		|		|
c		|		|
c		|		|
c		|		|
c		|    singly-    |
c		|    painted    |
c		|		|
c		|		|
c		|		|
c		|		|
c		|		|
c  itiedsplit =>|---------------|
c		|		|
c		|		|
c		|     ties      |
c           m =>|		|
c		|		|
c    numcells =>|_______________|
c
c  ihabs+1 column carries x-coord of cell
c  ihabs+2 column carries y-coord of cell
c  ihabs+3 column carries number of ties for this cell
c 
c
c  initialize II array with background value for re-use as output map
       do j=0,N
         do i=0,N
           II(i,j)=32767
         enddo
       enddo
c
c      set paint thresholds
       print *,"Calculating paint thresholds"
       do i=0,ihabs			! over all map categories
c
cc        don't bother if no cells requested of this cat
         if (prob(i) .eq. 0.0) goto 33
c
c         istart = 0
c         print *, "numcells is ", numcells
c         print *, "istart is ", istart
c         print *, "i is ", i
c         Call HeapSort(P,numcells,istart,i)	! sort P matrix by ith column
c
c
c        next category to be painted goes to single vector for sorting
         idispvec(0) = 0
         do l = 1, numcells
           idispvec(l) = P(l,i)
         enddo
c
c        call special vector version of Heapsort
         istart = 1
         Call HS2(idispvec,numcells+1,istart)
c
cc        diagnostic print
c         do k=16636,16646
c           print *, i, k, idispvec(k)
c         enddo
c
c       set paint threshold for this map category
         hslice(i) = idispvec(numcells - (int((prob(i) * 
     &    numcells)+.499)))
c       for absolute constraint maps
         if (hslice(i) .eq. 32767 .and. constraint(i)) hslice(i) = 32766
         print *, "trying to assign ", int((prob(i) *
     &    numcells)+.499), " cells for category", i
         jsum = jsum + int((prob(i) * numcells)+.499)
33     enddo                            ! over all map categories
c
       print*, 'trying to assign a total of', jsum,' cells of',
     &  numcells, ' total cells'
cc
cc       set paint threshold for this map category
c         hslice(i) = P(numcells - (int((prob(i) * numcells)+.499)),i)
c         print *, "trying to assign ", int((prob(i) *
c     &    numcells)+.499), " cells for category", i
c       enddo				! over all map categories
c
c      dump hslice array for checking
       print *, 'horizontal slices at these elevations:'
       do i=0,ihabs
         print *, i, hslice(i)
       enddo
c
c        write XPM file of fractal probability terrain aspect
       if (probout) then
         do iz = 0, ihabs
           if (constraint(iz) .or. prob(iz) .eq. 0.0 .or.
     &      iz .eq. igreatest) goto 335
           Call XPMRelief(P,iz,hslice(iz))
335      enddo
       endif
c
c
         numpainted = 0
         numtied = 0
         numblank = 0
c
       print *,"Beginning tie resolution process"
       do l = 1, numcells		! start at top of P matrix
         idummy = 0
         do j = 0, ihabs		! over all map categories
c          probably don't need to do this if prob(j) .eq. 0.0
           if (prob(j) .eq. 0.0 .or. j .eq. igreatest) goto 577
c
c         if this cell is painted this category
           if ( P(l,j) .gt. hslice(j) ) then
c           if ( P(l,j) .ge. hslice(j) ) then
c
c           count the multiple assignments
             P(l,ihabs+3) = P(l,ihabs+3) + 1
c
c           find the winner category for this cell
c           winner category has the highest probability surface
             if ( P(l,j) .gt. idummy ) then
               ihighest = j
               idummy = P(l,ihighest)
             endif
c
           endif			! this cell painted this category
c
577      enddo				! over all map categories
c         print *, "highest for cell ", l," was category", ihighest
c
c        keep count of each type of cell for displacement into
c         sorted array later
c        if singly-painted or currently tied, paint outmap with ihighest
         if ( P(l,ihabs+3) .eq. 1) then
c          paint this cell with the winning category
c           print *, "painting uncontested cell", P(l,ihabs+1),
c     &      P(l,ihabs+2)," with cat", ihighest
c
           numpainted = numpainted + 1
           II( P(l,ihabs+1),P(l,ihabs+2) ) = ihighest
         else if ( P(l,ihabs+3) .eq. 0) then
           numblank = numblank + 1
         else
           numtied = numtied + 1
c           print *, "painting tied cell", P(l,ihabs+1),P(l,ihabs+2),
c     &      " with winning cat", ihighest
           II( P(l,ihabs+1),P(l,ihabs+2) ) = ihighest
         endif
c
       enddo
c
c
c      at this point, II array has all winners and uncontested
c      cells painted
c
c      sort P array into 3 sections by number of assignments
c      ascending sort results in empty, singly-painted, and
c      ties sections
       print *,"Preparing to sort ties for resolution"
c      
c      sort by ties column
       istart = 1
       Call HeapSort(P,numcells,istart,ihabs+3)
c       print *, "got back from HeapSort"
c
c      calculate pointers to section breaks
       mtsplit = numblank
       itiedsplit = numblank + numpainted
c
       mtresolved = mtsplit
c
       print *, "numblank is ", numblank
       print *, "numtied is ", numtied
       print *, "itiedsplit is ", itiedsplit
       print *, "numpainted is ", numpainted
       print *, "mtresolved and mtsplit are ", mtresolved
c
       if (numtied .gt. numblank) print*, 'WARNING: More tied cells',
     &  ' than leftover blank cells to use for tie resolution!'
c
c      sum multiwayness of ties and compare to number of empty cells
       do m = itiedsplit+1, numcells
         isum = isum + ( P(m,ihabs+3) - 1 )
       enddo
       print *, "sum of multiwayness of ties is ", isum
       if (isum .gt. numblank) print*, 'Warning: cells are tied more',
     &  ' ways than there are blank cells to assign!'
c
c       call Print(P)
c
c  sort ties into priority order for resolution
c   by calculating and sorting on summed elevation 
c   for all tied categories
c
       do m = itiedsplit+1, numcells	! over all tied
         kount = 0
         idummy = 0
         do k = 0, ihabs                ! over all categories
c
c          probably not necessary if prob(k) .eq. 0
           if (prob(k) .eq. 0.0 .or. k .eq. igreatest) goto 933
c          if this cell is painted this category
           if ( P(m,k) .gt. hslice(k) ) then
c           if ( P(m,k) .ge. hslice(k) ) then
             kount = kount + 1          ! number of ties for this cell
             ties(kount) = k            ! keep categories of ties
c
           endif
933      enddo
c
         zmean = 0.0
         do k = 1, kount
           zmean = zmean + P(m,ties(k))
         enddo
c         P(m,ihabs+3) = zmean/kount
c        try weighting priority by simple sum, not average
c        therefore, max sum = (ihabs-1) * 32767
c        rescale as proportion of 32767 for integer*2 array cell
         P(m,ihabs+3) = zmean/(ihabs-1)
c
       enddo
c
c      sort ties by mean priority
       Call HeapSort(P,numtied,itiedsplit+1,ihabs+3)
c
c      save min and max tie priorities for gray levels in XPMTies
       maxgray = P(numcells,ihabs+3)
       print*, "maxgray is ", maxgray
c
c      generate xpm picture of tied cells
c       colored by priority gray shades
c       Call XPMTies(P)
c
c
c      start at bottom of list (highest priority ties)
       do m = numcells, itiedsplit+1, -1	! over all tied
         kount = 0
         idummy = 0
         do k = 0, ihabs		! over all categories
c
c          probably not necessary if prob(k) .eq. 0
           if (prob(k) .eq. 0.0 .or. k .eq. igreatest) goto 656
c          if this cell is painted this category
           if ( P(m,k) .gt. hslice(k) ) then
c           if ( P(m,k) .ge. hslice(k) ) then
             kount = kount + 1		! number of ties for this cell
             ties(kount) = k		! keep categories of ties
c            save most likely category for this cell
             if ( P(m,k) .gt. idummy ) then
               ihighest = k
               idummy = P(m,ihighest)
             endif
c
           endif
c
656      enddo
c
c
c        resolve all ties -
c        over all tied categories that are not the winner
c         (which has already been painted)
c         we need an empty cell -
c         search the empty cell list by this cat to find next
c         most likely available empty cell
c
c        empty cell list will be searched once for each tie category
c         except the winning category
c
c         print *, "deciding among ", kount, " ties for cell ", m
         do k = 1, kount		! over all ties for this cell
           if ( ties(k) .eq. ihighest ) goto 992	! don't re-do the winner
c           print *, ties(k), " was a tied category for cell ", m
c
           ibiggest = 0
           idummy = 0
           do i=1,mtresolved
             if(P(i,ties(k)) .gt. idummy) then
               ibiggest = i
               idummy = P(i,ties(k))
             endif
           enddo
c
           Call Interchange(P,ibiggest,mtresolved)
c
c           istart = 1
c           Call HeapSort(P,mtresolved,istart,ties(k))
c
c          now mtresolved points to highest
c          next available empty cell of category ties(k)
c          paint it!
c           print *, "painting empty cell", mtresolved," at coords ",
c     &      P(mtresolved,ihabs+1), P(mtresolved,ihabs+2)," with prob "
c     &      ,P(mtresolved,ties(k))," with cat", ties(k),
c     &      " for tie at cell", m
           II( P(mtresolved,ihabs+1),P(mtresolved,ihabs+2) ) = 
     &      ties(k)
c
c          burn one empty
           mtresolved = mtresolved - 1
c
992      enddo		! over all ties for this cell
c         print *, "resolving ties at cell ", m
c
       enddo		! over all tied cells
c
c**********************************************************************
c
c
c       Call Print(P)
c
c      generate xpm picture of tied cells
c       colored by priority gray shades
c       and used blank cells colored in yellow
       Call XPMTies(P)
c
c
c  set background category and
c  perform summary of realized map
       do i = 0,N
         do j = 0,N
           if (II(i,j) .eq. 32767) II(i,j) = igreatest
           kcnt(II(i,j)) = kcnt(II(i,j)) + 1
c          write(12,*) II(i,j)
         enddo
       enddo
c
c      if requested,
c       write GRASS ascii map
       if (grassout) then
         write(10,91), N
91       format ('north: ',i4)
         write(10,92)
92       format ('south: 0')
         write(10,93), N
93       format ('east: ',i4)
         write(10,94)
94       format ('west: 0')
         write(10,95), N+1
95       format ('rows: ',i4)
         write(10,96), N+1
96       format ('cols: ',i4)
         do i = 0,N
c           write(10,47) (II(i,j), j=0,N)
           write(10,47) (II(j,i), j=0,N)
47         format (<j>(i5,1x))
         enddo
       endif
c
c      write final map to xpm file
       if (xpmout) Call XPMMap(N,II)
c
       type 299
299    format (/t10,'Actual Map Summary:'/t12,'Habitat',5x,
     &         'P',3x,'Real Cells',3x,'Real P')
c       pfract = 0.0
       do i = 0, ihabs
          pfract = float(kcnt(i)) / float(numcells)
          type 300, i, prob(i), kcnt(i), pfract
300       format (t10, i6, 4x, f6.5, 2x, i6, 2x, f6.5)
       enddo
c
c**********************************************************************
c
c      visualize the synthetic landscape
       if (vizout) then
c
c  print final map to screen
         do i=0,N
           do j=0,N
             idispvec(j) = II(i,j)
           enddo
c           type 199, (idispvec(j), j=0,N)
c199         format (<N+1>(i1, 1x))
         enddo
c
c  print current state of P array
c         Call print(P)
c
c        visualize II array
         mapcolor = 'bobs.cmap\0'
         maptitle = 'Fractal Landscape Realizer \0'
         idtype = 4
         idumm = 0
         ipixid = -1
c
c   build display vector
c  make small maps more visible by repeating cells and lines
c         irepeat = maxN / N
c         print *, irepeat
         k=0
         do i=0,N
c           kstart = k
           do j=0,N
c             do kk = 1, irepeat
c               idispvec(k)=II(i,j)
               idispvec(k)=II(j,i)
               k=k+1
c             enddo	! repeating cells in this row
           enddo		! all cells in this row
c           kend = k-1
cc          repeat lines
c           if (irepeat .ge. 2) then
c             do jj = 1,irepeat-1
c               do kk = kstart, kend
c                 idispvec(k) = idispvec(kk)
c                 k = k+1
c               enddo	! dup as many times as necessary
c             enddo	! over all repeated lines
c           endif
c
         enddo
c
cc  adjust rows and columns by multiplier
c         isize = (N+1) * irepeat
c         jsize = (N+1) * irepeat
c         print *, isize, jsize
c         ierr=DrawPixmap(ipixid, isize, jsize, idtype, idispvec,
         ierr=DrawPixmap(ipixid,N+1,N+1,idtype,idispvec,
     &    idumm,mapcolor,maptitle)
c         print *,'ierr is', ierr
         print *, 'done visualizing'
         Call sleep(500)
c
       endif	! if visualizing landscape
c
       stop
       end
c
c **********************************************************************
c
       subroutine print(P)
       implicit real*4 (a-h,o-z)
       parameter (maxN=515)
       parameter (maxsp=10)
       integer*2 P(maxN*maxN,0:maxsp+3)
       integer*2 temp(maxsp+3)
c
       common/blkihab / ihabs,numcells,itiedsplit,
     &  irowcols,mtsplit,mtresolved,isupply,maxgray
c
       do i=1,numcells
          do j = 0,ihabs+3
            temp(j) = P(i,j)
          enddo
          type 188, i, (temp(j), j=0,ihabs+3)
c188       format (i3, 1x, <ihabs+3+1>(f4.2, 1x))
188       format (i5, 1x, <ihabs+1>(i5, 1x), 1x, 3(i5, 1x))
       enddo
c
       return
       end
c
c **********************************************************************
c
       subroutine mpfm2d (X, maxlevel, sigma, H, addition, iseed, wrap,
     &  gradient,iz)
       implicit real*4 (a-h,o-z)
       parameter (maxN=515)
       parameter (maxsp=10)
       real*4 X(0:maxN,0:maxN), H(maxN,maxsp)
       real*4 northeast, southeast, southwest, northwest
       integer*2 N, maxlevel, DD, d
       integer*4 iseed
       logical*1 addition, wrap, gradient
c
       common/blkcrnrs/ northeast, southeast, southwest, northwest
c
c  statment function definition
       f3(delta,x0,x1,x2) = (x0+x1+x2) / 3 + delta * Gauss(0,sigma)
       f4(delta,x0,x1,x2,x3) = (x0+x1+x2+x3) / 4 + delta *
     &   Gauss(0,sigma)
c
       print *, "Generating fractal"
       N = 2**maxlevel
       delta = sigma
       X(0,0) = delta * Gauss(0,sigma)
       X(0,N) = delta * Gauss(0,sigma)
       X(N,0) = delta * Gauss(0,sigma)
       X(N,N) = delta * Gauss(0,sigma)
c
       if (gradient) then
         X(0,0) = southwest
         X(0,N) = southeast
         X(N,0) = northwest
         X(N,N) = northeast
       endif
c
       DD = N
       d = N / 2
c
c*************************************************************************
       do is = 1, maxlevel
c  going from grid type I to type II
         delta = delta * 0.5**(0.5 * H(is,iz))
         do ix = d, N-d, DD
           do iy = d, N-d, DD
             X(ix,iy) = f4(delta, X(ix+d,iy+d), X(ix+d,iy-d), 
     &        X(ix-d,iy+d), X(ix-d,iy-d))
           enddo
         enddo
c  displace other points also if needed
         if (addition) then
           do ix = 0, N, DD
             do iy = 0, N, DD
               X(ix,iy) = X(ix,iy) + delta * Gauss(0,sigma)
             enddo
           enddo
         endif
c  going from grid type II to type I
         delta = delta * 0.5**(0.5*H(is,iz))
c  interpolate and offset boundary grid points
         do ix = d, N-d, DD
           X(ix,0) = f3(delta, X(ix+d,0), X(ix-d,0), X(ix,d))
           X(ix,N) = f3(delta, X(ix+d,N), X(ix-d,N), X(ix,N-d))
           X(0,ix) = f3(delta, X(0,ix+d), X(0,ix-d), X(d,ix))
           X(N,ix) = f3(delta, X(N,ix+d), X(N,ix-d), X(N-d,ix))
           if (wrap) then
             X(ix,N) = X(ix,0)
             X(N,ix) = X(0,ix)
           endif
         enddo
c  interpolate and offset inter grid points
         do ix = d, N-d, DD
           do iy = DD, N-d, DD
              X(ix,iy) = f4 (delta, X(ix,iy+d), X(ix,iy-d), 
     &         X(ix+d,iy), X(ix-d,iy))
           enddo
         enddo
         do ix = DD, N-d, DD
           do iy = d, N-d, DD
             X(ix,iy) = f4(delta, X(ix,iy+d), X(ix,iy-d), 
     &        X(ix+d,iy), X(ix-d,iy))
           enddo
         enddo
c  displace other points also if needed
         if (addition) then
           do ix = 0, N, DD
             do iy = 0, N, DD
               X(ix,iy) = X(ix,iy) + delta * Gauss(0,sigma)
             enddo
           enddo
           do ix = d, N-d, DD
             do iy = d, N-d, DD
               X(ix,iy) = X(ix,iy) + delta * Gauss(0,sigma)
             enddo
           enddo
         endif
c
         DD = DD / 2
         d = d / 2
c
       enddo	! for all levels
c*************************************************************************
c
       return
       end

c
c **********************************************************************
c
       function Gauss (xm, xs)
       implicit real*4 (a-h,o-z)
       real*4 p, xm, xs, sum
       integer*4 iseed
       common/blksd/ iseed
c
c create a normal number by summing 12 uniform random numbers
       sum = 0.0
       do i = 1,12
           p = ran(iseed)
           sum = sum + p
       enddo
       Gauss = (sum - 6.0) * xs + xm
c
       return
       end
c
c
c **********************************************************************
c
       subroutine StackSort(n,arr)
       dimension arr(n)
c
c  pick out each element in turn
       do 12 j=2,n
         a=arr(j)
c
c  look for the place to insert it
         do 11 i=j-1,1,-1
           if(arr(i).le.a)go to 10
           arr(i+1)=arr(i)
11       continue
         i=0
c
c  insert the number here
10       arr(i+1)=a
12     continue
       return
       end
c**********************************************************************
c
c  sort routine from Numerical Recipes
       subroutine sort (n, ra)
       dimension ra(n)
       l = n/2 + 1
       ir = n
c
10     continue
          if (l .gt. 1) then
             l = l-1
             rra = ra(l)
          else
             rra = ra(ir)
             ra(ir) = ra(1)
             ir = ir-1
             if (ir .eq. 1) then
                ra(1) = rra
                return
             endif
          endif
          i = l
          j = l+l
c
20     if (j .le. ir) then
           if (j .lt. ir) then
              if (ra(j) .lt. ra(j+1)) j = j+1
           endif
           if (rra .lt. ra(j)) then
               ra(i) = ra(j)
               i = j
               j = j+j
           else
               j = ir+1
           endif
           goto 20
       endif
c
       ra(i) = rra
       goto 10
c
       end
c**********************************************************************
       subroutine SwapDown(Heap,r,isize,istart,keyfield)
       implicit real*4 (a-h,o-z)
       parameter (maxN=515)
       parameter (maxsp=10)
       integer*2 Heap(maxN*maxN,0:maxsp+3)
       integer*2 child,r,isize,istart,keyfield,ipos
       logical*1 done
c
       common/blkihab / ihabs,numcells,itiedsplit,
     &  irowcols,mtsplit,mtresolved,isupply,maxgray
c
c       print *, 'entering SwapDown'
c       print *, 'params are ', r, isize, istart, keyfield
       ipos = istart - 1
       done = .false.
       child = 2 * r
       do while (.not. done .and. child .le. isize)
c  find the largest child
         if (child .lt. isize .and. Heap(child+ipos,keyfield) .lt.
     &   Heap(child+ipos+1,keyfield)) child = child + 1
c  interchange node and largest child if necessary
c   and move down to the next subtree
         if (Heap(r+ipos,keyfield) .lt. Heap(child+ipos,keyfield))
     &    then
           Call Interchange(Heap,r+ipos,child+ipos)
           r = child
           child = 2 * child
         else
           done = .true.
         endif
c
       enddo
c
       return
       end
c
c**********************************************************************
       subroutine Interchange(X,m,n)
       implicit real*4 (a-h,o-z)
       parameter (maxN=515)
       parameter (maxsp=10)
       integer*2 X(maxN*maxN,0:maxsp+3)
       integer*2 temp
c
       common/blkihab / ihabs,numcells,itiedsplit,
     &  irowcols,mtsplit,mtresolved,isupply,maxgray
c
       do i = 0, ihabs+3
         temp = X(m,i)
         X(m,i) = X(n,i)
         X(n,i) = temp
       enddo
c
       return
       end
c
c**********************************************************************
       subroutine HeapSort(X,isize,istart,keyfield)
       implicit real*4 (a-h,o-z)
       parameter (maxN=515)
       parameter (maxsp=10)
       integer*2 X(maxN*maxN,0:maxsp+3)
       integer*2 isize,r,tempr,keyfield,istart
c
       common/blkihab / ihabs,numcells,itiedsplit,
     &  irowcols,mtsplit,mtresolved,isupply,maxgray
c
c       print *, "isize in HeapSort is ", isize
c       print *, "istart in HeapSort is ", istart
c       print *, "keyfield in HeapSort is ", keyfield
c      heapify
       do r = isize/2, 1, -1
         tempr = r
         Call SwapDown(X,r,isize,istart,keyfield)
         r = tempr
       enddo
c
c       print*, "P after heapify is "
c       do i=istart,isize
c         print *, i, X(i,keyfield)
c       enddo
c
c   repeatedly put largest item in root at end of list,
c   prune it from the tree
c   and apply SwapDown to the rest of the tree
c
       do i = isize, 2, -1
         Call Interchange(X,istart,(i+istart-1))
c         r = isize
         r = 1
         Call SwapDown(X,r,i-1,istart,keyfield)
c
c         print*, "P after Swapdown is  "
c         do j=istart,isize
c           print *, j, X(j,keyfield)
c         enddo

       enddo
c
       return
       end
c
c**********************************************************************
       Subroutine XPMMap(N,II)
c
c  output routine for xpm X-window display tool
c
c
       implicit real*4 (a-h,o-z)
       parameter (maxN=515)
       parameter (maxsp=10)
       integer*2 N, II(0:maxN,0:maxN)
       integer*2 itemp(0:maxN)
c
c
c  output map
       print *, "writing final map to xpm format file landscape.xpm"
c
       open (20,file='landscape.xpm',status='unknown')
c
       write (20, 100)
100    format ('/* XPM */',/'static char *realizer[]',
     & ' = {',/'/* cols rows ncolors chars_per_pixel',
     & ' */')
       write (20, 102) N+1, N+1
102    format ('"',i6,1x,i6,1x,'11 1 0 0",',/'/* colors */')
       write (20, 104)
104    format ('"0  c white    m white',
     & '",',/'"1  s Decid  c BurlyWood    m black',
     & '",',/'"2  s Mixed  c DarkGoldenrod3       m gray40',
     & '",',/'"3  s Everg  c MistyRose3         m gray50',
     & '",',/'"4  s Trans  c LightGoldenrod4         m gray60',
     & '",',/'"5  s Urban  c DarkGoldenrod1       m gray70',
     & '",',/'"6  s Water  c gold    m gray80',
     & '",',/'"7  s Maint  c LightGoldenrod3       m white',
     & '",',/'"8  s Cropl  c SaddleBrown    m lightgray',
     & '",',/'"9  s Lawng  c RosyBrown    m gray',
     & '",',/'"*  c chocolate     m black",',/'/* pixels */')
c
       do i= 0, N
          do j = 0, N
c            itemp(j) = II(i,j)
            itemp(j) = II(j,i)
          enddo
          write (20,108) (itemp(j), j=0,N)
108       format (' "',<N+1>(i1),'",')
       enddo
c
c
       write (20,112)
112    format ('};')
c
       return
       end
c
c**********************************************************
      Subroutine InputMap(mapfile,P,N,k,invert)
c
c   input map from a data file
c
       implicit real*4 (a-h,o-z)
       parameter (maxN=515)
       parameter (maxsp=10)
       integer*2 P(maxN*maxN,0:maxsp+3)
       integer*2 N, k
       real*4 X(0:maxN,0:maxN)
       character*60 mapfile
       logical*1 invert
c
       common/blkihab / ihabs,numcells,itiedsplit,
     &  irowcols,mtsplit,mtresolved,isupply,maxgray
c
       open (14, file=mapfile, status='old')
c
       xmin = 1e30
       xmax = -1e30
       do j = 0, N
           read (14,*) ( X(i,j), i = 0, N)
c           read (14,118) X(i,j)
c118        format (<(N+1)*(N+1)>i5, 1x)
         do i = 0, N
           if (X(i,j) .lt. xmin) xmin = X(i,j)
           if (X(i,j) .gt. xmax) xmax = X(i,j)
         enddo
       enddo
c
c      invert if necessary and scale between 0 and 32767
       xrange = xmax - xmin
       do j = 0, N
         do i = 0, N
           if (invert) then
             temp = ((-X(i,j) + xmax) / xrange) * 32767
           else
             temp = ((X(i,j) - xmin) / xrange) * 32767
           endif
c
           iv = (i*(N+1))+j+1
           P(iv,k) = nint(temp)
           P(iv,ihabs+1) = i
           P(iv,ihabs+2) = j
c           print *, 'row ',i,' col ', j
         enddo
       enddo
c
       close(14)
       rewind(14)
c
       return
       end
c**********************************************************
       subroutine SD2(Heap,r,isize,istart)
       implicit real*4 (a-h,o-z)
       parameter (maxN=515)
       parameter (maxsp=10)
       integer*2 Heap(maxN*maxN)
       integer*2 child,r,isize,istart,ipos
       logical*1 done
c
       common/blkihab / ihabs,numcells,itiedsplit,
     &  irowcols,mtsplit,mtresolved,isupply,maxgray
c
c       print *, 'entering SwapDown'
c       print *, 'params are ', r, isize, istart
       ipos = istart - 1
       done = .false.
       child = 2 * r
       do while (.not. done .and. child .le. isize)
c  find the largest child
         if (child .lt. isize .and. Heap(child+ipos) .lt.
     &   Heap(child+ipos+1)) child = child + 1
c  interchange node and largest child if necessary
c   and move down to the next subtree
         if (Heap(r+ipos) .lt. Heap(child+ipos))
     &    then
           Call I2(Heap,r+ipos,child+ipos)
           r = child
           child = 2 * child
         else
           done = .true.
         endif
c
       enddo
c
       return
       end
c
c**********************************************************************
       subroutine I2(X,m,n)
       implicit real*4 (a-h,o-z)
       parameter (maxN=515)
       parameter (maxsp=10)
       integer*2 X(maxN*maxN)
       integer*2 temp
c
       common/blkihab / ihabs,numcells,itiedsplit,
     &  irowcols,mtsplit,mtresolved,isupply,maxgray
c
       temp = X(m)
       X(m) = X(n)
       X(n) = temp
c
       return
       end
c
c**********************************************************************
       subroutine HS2(X,isize,istart)
       implicit real*4 (a-h,o-z)
       parameter (maxN=515)
       parameter (maxsp=10)
       integer*2 X(maxN*maxN)
       integer*2 isize,r,tempr,istart
c
       common/blkihab / ihabs,numcells,itiedsplit,
     &  irowcols,mtsplit,mtresolved,isupply,maxgray
c
c       print *, "isize in HeapSort is ", isize
c       print *, "istart in HeapSort is ", istart
c      heapify
       do r = isize/2, 1, -1
         tempr = r
         Call SD2(X,r,isize,istart)
         r = tempr
       enddo
c
c       print*, "P after heapify is "
c       do i=istart,isize
c         print *, i, X(i)
c       enddo
c
c   repeatedly put largest item in root at end of list,
c   prune it from the tree
c   and apply SwapDown to the rest of the tree
c
       do i = isize, 2, -1
         Call I2(X,istart,(i+istart-1))
c         r = isize
         r = 1
         Call SD2(X,r,i-1,istart)
c
c         print*, "P after Swapdown is  "
c         do j=istart,isize
c           print *, j, X(j)
c         enddo

       enddo
c
       return
       end
c
c**********************************************************************
       Subroutine XPMTies(P)
c
c  output routine for xpm X-window display tool
c
c  write xpm format map with grayshades
c
       parameter (maxN=515)
       parameter (maxsp=10)
       implicit real*4 (a-h,o-z)
       integer*2 P(maxN*maxN,0:maxsp+3)
       integer*2 itemp(0:maxN,0:maxN)
c
       common/blkihab / ihabs,numcells,itiedsplit,
     &  irowcols,mtsplit,mtresolved,isupply,maxgray
c
       open (40,file='ties.xpm',status='unknown')
c
       type 99
99     format ('XPM map file ties.xpm being written')
       write (40, 100)
100    format ('/* XPM */',/'static char *tiesmap[]',
     & ' = {',/'/* cols rows ncolors chars_per_pixel',
     & ' */')
       write (40, 102) irowcols+1, irowcols+1
102    format (' "',i6,1x,i6,1x,'108 3 0 0",',/'/* colors */')
       write (40, 104)
104    format ('"  0 s 0.0prob c white g4 white g white m white',
     & '",',/'"  1 s 0.01prob c gray99 g4 gray99 g gray99 m gray99',
     & '",',/'"  2 s 0.02prob c gray98 g4 gray98 g gray98 m gray98',
     & '",',/'"  3 s 0.03prob c gray97 g4 gray97 g gray97 m gray97',
     & '",',/'"  4 s 0.04prob c gray96 g4 gray96 g gray96 m gray96',
     & '",',/'"  5 s 0.05prob c gray95 g4 gray95 g gray95 m gray95',
     & '",',/'"  6 s 0.06prob c gray94 g4 gray94 g gray94 m gray94',
     & '",',/'"  7 s 0.07prob c gray93 g4 gray93 g gray93 m gray93',
     & '",',/'"  8 s 0.08prob c gray92 g4 gray92 g gray92 m gray92',
     & '",',/'"  9 s 0.09prob c gray91 g4 gray91 g gray91 m gray91',
     & '",',/'" 10 s 0.10prob c gray90 g4 gray90 g gray90 m gray90',
     & '",',/'" 11 s 0.11prob c gray89 g4 gray89 g gray89 m gray89',
     & '",',/'" 12 s 0.12prob c gray88 g4 gray88 g gray88 m gray88',
     & '",',/'" 13 s 0.13prob c gray87 g4 gray87 g gray87 m gray87',
     & '",',/'" 14 s 0.14prob c gray86 g4 gray86 g gray86 m gray86',
     & '",',/'" 15 s 0.15prob c gray85 g4 gray85 g gray85 m gray85',
     & '",',/'" 16 s 0.16prob c gray84 g4 gray84 g gray84 m gray84',
     & '",',/'" 17 s 0.17prob c gray83 g4 gray83 g gray83 m gray83',
     & '",',/'" 18 s 0.18prob c gray82 g4 gray82 g gray82 m gray82',
     & '",',/'" 19 s 0.19prob c gray81 g4 gray81 g gray81 m gray81',
     & '",',/'" 20 s 0.20prob c gray80 g4 gray80 g gray80 m gray80',
     & '",',/'" 21 s 0.21prob c gray79 g4 gray79 g gray79 m gray79',
     & '",',/'" 22 s 0.22prob c gray78 g4 gray78 g gray78 m gray78',
     & '",',/'" 23 s 0.23prob c gray77 g4 gray77 g gray77 m gray77',
     & '",',/'" 24 s 0.24prob c gray76 g4 gray76 g gray76 m gray76',
     & '",',/'" 25 s 0.25prob c gray75 g4 gray75 g gray75 m gray75',
     & '",',/'" 26 s 0.26prob c gray74 g4 gray74 g gray74 m gray74',
     & '",',/'" 27 s 0.27prob c gray73 g4 gray73 g gray73 m gray73',
     & '",',/'" 28 s 0.28prob c gray72 g4 gray72 g gray72 m gray72',
     & '",',/'" 29 s 0.29prob c gray71 g4 gray71 g gray71 m gray71',
     & '",',/'" 30 s 0.30prob c gray70 g4 gray70 g gray70 m gray70',
     & '",',/'" 31 s 0.31prob c gray69 g4 gray69 g gray69 m gray69',
     & '",',/'" 32 s 0.32prob c gray68 g4 gray68 g gray68 m gray68',
     & '",',/'" 33 s 0.33prob c gray67 g4 gray67 g gray67 m gray67',
     & '",',/'" 34 s 0.34prob c gray66 g4 gray66 g gray66 m gray66',
     & '",',/'" 35 s 0.35prob c gray65 g4 gray65 g gray65 m gray65',
     & '",',/'" 36 s 0.36prob c gray64 g4 gray64 g gray64 m gray64',
     & '",',/'" 37 s 0.37prob c gray63 g4 gray63 g gray63 m gray63',
     & '",',/'" 38 s 0.38prob c gray62 g4 gray62 g gray62 m gray62',
     & '",',/'" 39 s 0.39prob c gray61 g4 gray61 g gray61 m gray61',
     & '",',/'" 40 s 0.40prob c gray60 g4 gray60 g gray60 m gray60',
     & '",',/'" 41 s 0.41prob c gray59 g4 gray59 g gray59 m gray59',
     & '",',/'" 42 s 0.42prob c gray58 g4 gray58 g gray58 m gray58',
     & '",',/'" 43 s 0.43prob c gray57 g4 gray57 g gray57 m gray57',
     & '",',/'" 44 s 0.44prob c gray56 g4 gray56 g gray56 m gray56',
     & '",',/'" 45 s 0.45prob c gray55 g4 gray55 g gray55 m gray55',
     & '",',/'" 46 s 0.46prob c gray54 g4 gray54 g gray54 m gray54',
     & '",',/'" 47 s 0.47prob c gray53 g4 gray53 g gray53 m gray53',
     & '",',/'" 48 s 0.48prob c gray52 g4 gray52 g gray52 m gray52',
     & '",',/'" 49 s 0.49prob c gray51 g4 gray51 g gray51 m gray51',
     & '",',/'" 50 s 0.50prob c gray50 g4 gray50 g gray50 m gray50',
     & '",',/'" 51 s 0.51prob c gray49 g4 gray49 g gray49 m gray49',
     & '",',/'" 52 s 0.52prob c gray48 g4 gray48 g gray48 m gray48',
     & '",',/'" 53 s 0.53prob c gray47 g4 gray47 g gray47 m gray47',
     & '",',/'" 54 s 0.54prob c gray46 g4 gray46 g gray46 m gray46',
     & '",',/'" 55 s 0.55prob c gray45 g4 gray45 g gray45 m gray45',
     & '",',/'" 56 s 0.56prob c gray44 g4 gray44 g gray44 m gray44',
     & '",',/'" 57 s 0.57prob c gray43 g4 gray43 g gray43 m gray43',
     & '",',/'" 58 s 0.58prob c gray42 g4 gray42 g gray42 m gray42',
     & '",',/'" 59 s 0.59prob c gray41 g4 gray41 g gray41 m gray41',
     & '",',/'" 60 s 0.60prob c gray40 g4 gray40 g gray40 m gray40',
     & '",',/'" 61 s 0.61prob c gray39 g4 gray39 g gray39 m gray39',
     & '",',/'" 62 s 0.62prob c gray38 g4 gray38 g gray38 m gray38',
     & '",',/'" 63 s 0.63prob c gray37 g4 gray37 g gray37 m gray37',
     & '",',/'" 64 s 0.64prob c gray36 g4 gray36 g gray36 m gray36',
     & '",',/'" 65 s 0.65prob c gray35 g4 gray35 g gray35 m gray35',
     & '",',/'" 66 s 0.66prob c gray34 g4 gray34 g gray34 m gray34',
     & '",',/'" 67 s 0.67prob c gray33 g4 gray33 g gray33 m gray33',
     & '",',/'" 68 s 0.68prob c gray32 g4 gray32 g gray32 m gray32',
     & '",',/'" 69 s 0.69prob c gray31 g4 gray31 g gray31 m gray31',
     & '",',/'" 70 s 0.70prob c gray30 g4 gray30 g gray30 m gray30',
     & '",',/'" 71 s 0.71prob c gray29 g4 gray29 g gray29 m gray29',
     & '",',/'" 72 s 0.72prob c gray28 g4 gray28 g gray28 m gray28',
     & '",',/'" 73 s 0.73prob c gray27 g4 gray27 g gray27 m gray27',
     & '",',/'" 74 s 0.74prob c gray26 g4 gray26 g gray26 m gray26',
     & '",',/'" 75 s 0.75prob c gray25 g4 gray25 g gray25 m gray25',
     & '",',/'" 76 s 0.76prob c gray24 g4 gray24 g gray24 m gray24',
     & '",',/'" 77 s 0.77prob c gray23 g4 gray23 g gray23 m gray23',
     & '",',/'" 78 s 0.78prob c gray22 g4 gray22 g gray22 m gray22',
     & '",',/'" 79 s 0.79prob c gray21 g4 gray21 g gray21 m gray21',
     & '",',/'" 80 s 0.80prob c gray20 g4 gray20 g gray20 m gray20',
     & '",',/'" 81 s 0.81prob c gray19 g4 gray19 g gray19 m gray19',
     & '",',/'" 82 s 0.82prob c gray18 g4 gray18 g gray18 m gray18',
     & '",',/'" 83 s 0.83prob c gray17 g4 gray17 g gray17 m gray17',
     & '",',/'" 84 s 0.84prob c gray16 g4 gray16 g gray16 m gray16',
     & '",',/'" 85 s 0.85prob c gray15 g4 gray15 g gray15 m gray15',
     & '",',/'" 86 s 0.86prob c gray14 g4 gray14 g gray14 m gray14',
     & '",',/'" 87 s 0.87prob c gray13 g4 gray13 g gray13 m gray13',
     & '",',/'" 88 s 0.88prob c gray12 g4 gray12 g gray12 m gray12',
     & '",',/'" 89 s 0.89prob c gray11 g4 gray11 g gray11 m gray11',
     & '",',/'" 90 s 0.90prob c gray10 g4 gray10 g gray10 m gray10',
     & '",',/'" 91 s 0.91prob c gray9 g4 gray9 g gray9 m gray9',
     & '",',/'" 92 s 0.92prob c gray8 g4 gray8 g gray8 m gray8',
     & '",',/'" 93 s 0.93prob c gray7 g4 gray7 g gray7 m gray7',
     & '",',/'" 94 s 0.94prob c gray6 g4 gray6 g gray6 m gray6',
     & '",',/'" 95 s 0.95prob c gray5 g4 gray5 g gray5 m gray5',
     & '",',/'" 96 s 0.96prob c gray4 g4 gray4 g gray4 m gray4',
     & '",',/'" 97 s 0.97prob c gray3 g4 gray3 g gray3 m gray3',
     & '",',/'" 98 s 0.98prob c gray2 g4 gray2 g gray2 m gray2',
     & '",',/'" 99 s 0.99prob c gray1 g4 gray1 g gray1 m gray1',
     & '",',/'"100 s 1.00prob c black g4 black g black m black',
     & '",',/'"101  c white    m white  s NotFlammable',
     & '",',/'"102  c GreenYellow       m gray70  s LP0',
     & '",',/'"103  c LawnGreen         m gray60 s LP1',
     & '",',/'"104  c LimeGreen         m gray50   s LP2',
     & '",',/'"105  c ForestGreen       m gray40   s LP3',
     & '",',/'"106  c yellow    m gray80  s NF',
     & '",',/'"107  c blue    m black  s water ",',
     & /'/* pixels */')
c
c
c      calculate a log scale for gray levels
c      use this scale log for uniform scaling across realizations
c       scalelog = 10**(log10(100.)/32767.)
c      scalelog for simple summed priority
c       scale max gray to ihabs-isupply-1 full-on categories
c        (= ihabs-isupply-1 x 32767)
cc       scalelog = 10**(log10(100.)/(32767. * (ihabs-isupply-1)))
c      use this scalelog for maximum priority resolution w/in one map
       scalelog = 10**(log10(100.) / float(maxgray))
c       print *, scalelog
c
c      start at bottom of list (highest priority ties)
       do m = numcells, itiedsplit+1, -1	! over all tied
c         itemp(P(m,ihabs+1),P(m,ihabs+2)) = 
c     &    (float(P(m,ihabs+3))/32767)*100
c
c         print *, scalelog**P(m,ihabs+3)
c         print *, P(m,ihabs+1),P(m,ihabs+2),P(m,ihabs+3)
         itemp(P(m,ihabs+1),P(m,ihabs+2)) = scalelog**P(m,ihabs+3)
c         itemp(P(m,ihabs+1),P(m,ihabs+2)) = 100
c        all priorities greater than color index 100 set black
         if (itemp(P(m,ihabs+1),P(m,ihabs+2)) .gt. 100)
     &    itemp(P(m,ihabs+1),P(m,ihabs+2)) = 100
       enddo
c
c      color used blanks yellow
       do m = mtsplit,mtresolved+1,-1
         itemp(P(m,ihabs+1),P(m,ihabs+2)) = 106
       enddo
c
       do i=0, irowcols
          write (40,108) (itemp(j,i), j=0, irowcols)
108       format ('"',<irowcols+1>(i3),'",')
       enddo
c
       write (40,112)
112    format ('};')
       close (40)
c
       return
       end
c
c**********************************************************
       Subroutine XPMRelief(P,iz,iflood)
c
c  output routine for xpm X-window display tool
c
c  write xpm format aspect map of painted fractal
c   spatial probability surfaces with grayshades
c
       parameter (maxN=515)
       parameter (maxsp=10)
       implicit real*4 (a-h,o-z)
       integer*2 P(maxN*maxN,0:maxsp+3)
       integer*2 itemp(0:maxN,0:maxN), idispvec(0:maxN)
       integer*2 iflood
       character*15 outfile(maxsp)
c
       common/blkihab / ihabs,numcells,itiedsplit,
     &  irowcols,mtsplit,mtresolved,isupply,maxgray
c
c      set filenames for the first 10 categories
       outfile(0) = 'probmap0.xpm   '
       outfile(1) = 'probmap1.xpm   '
       outfile(2) = 'probmap2.xpm   '
       outfile(3) = 'probmap3.xpm   '
       outfile(4) = 'probmap4.xpm   '
       outfile(5) = 'probmap5.xpm   '
       outfile(6) = 'probmap6.xpm   '
       outfile(7) = 'probmap7.xpm   '
       outfile(8) = 'probmap8.xpm   '
       outfile(9) = 'probmap9.xpm   '
       outfile(10) = 'probmap10.xpm  '
c
c
       type 99, outfile(iz)
99     format ('XPM map file ', a15 'being written')
c
c      open the right file for xpm output
       open (50,file=outfile(iz),status='unknown')
c
       write (50, 100)
100    format ('/* XPM */',/'static char *fracasp[]',
     & ' = {',/'/* cols rows ncolors chars_per_pixel',
     & ' */')
       write (50, 102) irowcols+1, irowcols+1
102    format ('"',i6,1x,i6,1x,'108 3 0 0",',/'/* colors */')
       write (50, 104)
104    format ('"  0 s 0.0prob c white g4 white g white m white',
     & '",',/'"  1 s 0.01prob c gray99 g4 gray99 g gray99 m gray99',
     & '",',/'"  2 s 0.02prob c gray98 g4 gray98 g gray98 m gray98',
     & '",',/'"  3 s 0.03prob c gray97 g4 gray97 g gray97 m gray97',
     & '",',/'"  4 s 0.04prob c gray96 g4 gray96 g gray96 m gray96',
     & '",',/'"  5 s 0.05prob c gray95 g4 gray95 g gray95 m gray95',
     & '",',/'"  6 s 0.06prob c gray94 g4 gray94 g gray94 m gray94',
     & '",',/'"  7 s 0.07prob c gray93 g4 gray93 g gray93 m gray93',
     & '",',/'"  8 s 0.08prob c gray92 g4 gray92 g gray92 m gray92',
     & '",',/'"  9 s 0.09prob c gray91 g4 gray91 g gray91 m gray91',
     & '",',/'" 10 s 0.10prob c gray90 g4 gray90 g gray90 m gray90',
     & '",',/'" 11 s 0.11prob c gray89 g4 gray89 g gray89 m gray89',
     & '",',/'" 12 s 0.12prob c gray88 g4 gray88 g gray88 m gray88',
     & '",',/'" 13 s 0.13prob c gray87 g4 gray87 g gray87 m gray87',
     & '",',/'" 14 s 0.14prob c gray86 g4 gray86 g gray86 m gray86',
     & '",',/'" 15 s 0.15prob c gray85 g4 gray85 g gray85 m gray85',
     & '",',/'" 16 s 0.16prob c gray84 g4 gray84 g gray84 m gray84',
     & '",',/'" 17 s 0.17prob c gray83 g4 gray83 g gray83 m gray83',
     & '",',/'" 18 s 0.18prob c gray82 g4 gray82 g gray82 m gray82',
     & '",',/'" 19 s 0.19prob c gray81 g4 gray81 g gray81 m gray81',
     & '",',/'" 20 s 0.20prob c gray80 g4 gray80 g gray80 m gray80',
     & '",',/'" 21 s 0.21prob c gray79 g4 gray79 g gray79 m gray79',
     & '",',/'" 22 s 0.22prob c gray78 g4 gray78 g gray78 m gray78',
     & '",',/'" 23 s 0.23prob c gray77 g4 gray77 g gray77 m gray77',
     & '",',/'" 24 s 0.24prob c gray76 g4 gray76 g gray76 m gray76',
     & '",',/'" 25 s 0.25prob c gray75 g4 gray75 g gray75 m gray75',
     & '",',/'" 26 s 0.26prob c gray74 g4 gray74 g gray74 m gray74',
     & '",',/'" 27 s 0.27prob c gray73 g4 gray73 g gray73 m gray73',
     & '",',/'" 28 s 0.28prob c gray72 g4 gray72 g gray72 m gray72',
     & '",',/'" 29 s 0.29prob c gray71 g4 gray71 g gray71 m gray71',
     & '",',/'" 30 s 0.30prob c gray70 g4 gray70 g gray70 m gray70',
     & '",',/'" 31 s 0.31prob c gray69 g4 gray69 g gray69 m gray69',
     & '",',/'" 32 s 0.32prob c gray68 g4 gray68 g gray68 m gray68',
     & '",',/'" 33 s 0.33prob c gray67 g4 gray67 g gray67 m gray67',
     & '",',/'" 34 s 0.34prob c gray66 g4 gray66 g gray66 m gray66',
     & '",',/'" 35 s 0.35prob c gray65 g4 gray65 g gray65 m gray65',
     & '",',/'" 36 s 0.36prob c gray64 g4 gray64 g gray64 m gray64',
     & '",',/'" 37 s 0.37prob c gray63 g4 gray63 g gray63 m gray63',
     & '",',/'" 38 s 0.38prob c gray62 g4 gray62 g gray62 m gray62',
     & '",',/'" 39 s 0.39prob c gray61 g4 gray61 g gray61 m gray61',
     & '",',/'" 40 s 0.40prob c gray60 g4 gray60 g gray60 m gray60',
     & '",',/'" 41 s 0.41prob c gray59 g4 gray59 g gray59 m gray59',
     & '",',/'" 42 s 0.42prob c gray58 g4 gray58 g gray58 m gray58',
     & '",',/'" 43 s 0.43prob c gray57 g4 gray57 g gray57 m gray57',
     & '",',/'" 44 s 0.44prob c gray56 g4 gray56 g gray56 m gray56',
     & '",',/'" 45 s 0.45prob c gray55 g4 gray55 g gray55 m gray55',
     & '",',/'" 46 s 0.46prob c gray54 g4 gray54 g gray54 m gray54',
     & '",',/'" 47 s 0.47prob c gray53 g4 gray53 g gray53 m gray53',
     & '",',/'" 48 s 0.48prob c gray52 g4 gray52 g gray52 m gray52',
     & '",',/'" 49 s 0.49prob c gray51 g4 gray51 g gray51 m gray51',
     & '",',/'" 50 s 0.50prob c gray50 g4 gray50 g gray50 m gray50',
     & '",',/'" 51 s 0.51prob c gray49 g4 gray49 g gray49 m gray49',
     & '",',/'" 52 s 0.52prob c gray48 g4 gray48 g gray48 m gray48',
     & '",',/'" 53 s 0.53prob c gray47 g4 gray47 g gray47 m gray47',
     & '",',/'" 54 s 0.54prob c gray46 g4 gray46 g gray46 m gray46',
     & '",',/'" 55 s 0.55prob c gray45 g4 gray45 g gray45 m gray45',
     & '",',/'" 56 s 0.56prob c gray44 g4 gray44 g gray44 m gray44',
     & '",',/'" 57 s 0.57prob c gray43 g4 gray43 g gray43 m gray43',
     & '",',/'" 58 s 0.58prob c gray42 g4 gray42 g gray42 m gray42',
     & '",',/'" 59 s 0.59prob c gray41 g4 gray41 g gray41 m gray41',
     & '",',/'" 60 s 0.60prob c gray40 g4 gray40 g gray40 m gray40',
     & '",',/'" 61 s 0.61prob c gray39 g4 gray39 g gray39 m gray39',
     & '",',/'" 62 s 0.62prob c gray38 g4 gray38 g gray38 m gray38',
     & '",',/'" 63 s 0.63prob c gray37 g4 gray37 g gray37 m gray37',
     & '",',/'" 64 s 0.64prob c gray36 g4 gray36 g gray36 m gray36',
     & '",',/'" 65 s 0.65prob c gray35 g4 gray35 g gray35 m gray35',
     & '",',/'" 66 s 0.66prob c gray34 g4 gray34 g gray34 m gray34',
     & '",',/'" 67 s 0.67prob c gray33 g4 gray33 g gray33 m gray33',
     & '",',/'" 68 s 0.68prob c gray32 g4 gray32 g gray32 m gray32',
     & '",',/'" 69 s 0.69prob c gray31 g4 gray31 g gray31 m gray31',
     & '",',/'" 70 s 0.70prob c gray30 g4 gray30 g gray30 m gray30',
     & '",',/'" 71 s 0.71prob c gray29 g4 gray29 g gray29 m gray29',
     & '",',/'" 72 s 0.72prob c gray28 g4 gray28 g gray28 m gray28',
     & '",',/'" 73 s 0.73prob c gray27 g4 gray27 g gray27 m gray27',
     & '",',/'" 74 s 0.74prob c gray26 g4 gray26 g gray26 m gray26',
     & '",',/'" 75 s 0.75prob c gray25 g4 gray25 g gray25 m gray25',
     & '",',/'" 76 s 0.76prob c gray24 g4 gray24 g gray24 m gray24',
     & '",',/'" 77 s 0.77prob c gray23 g4 gray23 g gray23 m gray23',
     & '",',/'" 78 s 0.78prob c gray22 g4 gray22 g gray22 m gray22',
     & '",',/'" 79 s 0.79prob c gray21 g4 gray21 g gray21 m gray21',
     & '",',/'" 80 s 0.80prob c gray20 g4 gray20 g gray20 m gray20',
     & '",',/'" 81 s 0.81prob c gray19 g4 gray19 g gray19 m gray19',
     & '",',/'" 82 s 0.82prob c gray18 g4 gray18 g gray18 m gray18',
     & '",',/'" 83 s 0.83prob c gray17 g4 gray17 g gray17 m gray17',
     & '",',/'" 84 s 0.84prob c gray16 g4 gray16 g gray16 m gray16',
     & '",',/'" 85 s 0.85prob c gray15 g4 gray15 g gray15 m gray15',
     & '",',/'" 86 s 0.86prob c gray14 g4 gray14 g gray14 m gray14',
     & '",',/'" 87 s 0.87prob c gray13 g4 gray13 g gray13 m gray13',
     & '",',/'" 88 s 0.88prob c gray12 g4 gray12 g gray12 m gray12',
     & '",',/'" 89 s 0.89prob c gray11 g4 gray11 g gray11 m gray11',
     & '",',/'" 90 s 0.90prob c gray10 g4 gray10 g gray10 m gray10',
     & '",',/'" 91 s 0.91prob c gray9 g4 gray9 g gray9 m gray9',
     & '",',/'" 92 s 0.92prob c gray8 g4 gray8 g gray8 m gray8',
     & '",',/'" 93 s 0.93prob c gray7 g4 gray7 g gray7 m gray7',
     & '",',/'" 94 s 0.94prob c gray6 g4 gray6 g gray6 m gray6',
     & '",',/'" 95 s 0.95prob c gray5 g4 gray5 g gray5 m gray5',
     & '",',/'" 96 s 0.96prob c gray4 g4 gray4 g gray4 m gray4',
     & '",',/'" 97 s 0.97prob c gray3 g4 gray3 g gray3 m gray3',
     & '",',/'" 98 s 0.98prob c gray2 g4 gray2 g gray2 m gray2',
     & '",',/'" 99 s 0.99prob c gray1 g4 gray1 g gray1 m gray1',
     & '",',/'"100 s 1.00prob c black g4 black g black m black',
     & '",',/'"101  c white    m white  s NotFlammable',
     & '",',/'"102  c GreenYellow       m gray70  s LP0',
     & '",',/'"103  c LawnGreen         m gray60 s LP1',
     & '",',/'"104  c LimeGreen         m gray50   s LP2',
     & '",',/'"105  c ForestGreen       m gray40   s LP3',
     & '",',/'"106  c yellow    m gray80  s NF',
     & '",',/'"107  c blue    m black  s water ",',
     & /'/* pixels */')
c
c
c      load fractal probability terrain into itemp array
       do m = 1, numcells       ! over all cells
         itemp(P(m,ihabs+1),P(m,ihabs+2)) = P(m,iz)
       enddo
c
c
c  equations for slope and aspect are from:
c  Horn, B.K.P., 1981.  Hill Shading and the Reflectance Map.
c       Proceedings of the I.E.E.E., 69(1):14-47.
c
c  calculate and display aspect map
c  set sun altitude above horizon in degrees (0 to 90)
       alt=40.
c  set sun azimuth in degrees west of north (-1 to 360)
       az=105.
       k=0
       do j=0,irowcols
         do i=0,irowcols
c
           x=(itemp(i-1,j-1)+2.*itemp(i,j-1)+itemp(i+1,j-1)
     &      -itemp(i-1,j+1)-2.*itemp(i,j+1)-itemp(i+1,j+1))
     &      /(8.0*30.0)
c
           y=(itemp(i-1,j-1)+2.*itemp(i-1,j)+itemp(i-1,j+1)
     &      -itemp(i+1,j-1)-2.*itemp(i+1,j)-itemp(i+1,j+1))
     &      /(8.0*30.0)
c
           slope=90.-atand(sqrt(x*x + y*y))
           if (x .eq. 0. .and. y .eq. 0.) then
             aspect=360.
           else
             aspect=atan2d(x,y)
           endif
c
           c=sind(alt)*sind(slope)+
     &      cosd(alt)*cosd(slope)*cosd(az-aspect)
c
           if (c .le. 0.) then
             idispvec(k)=0
           else
             idispvec(k)=nint(c*100.)
           endif
c
c          flood with water if not painted
           if (itemp(i,j) .le. iflood) idispvec(k) = 107
c
c           print *, i,j,k,idispvec(k)
           k=k+1
         enddo
         write (50,108) (idispvec(l), l=0, irowcols)
108      format (' "',<irowcols+1>(i3),'",')
         k=0
       enddo
c
       write (50,112)
112    format ('};')
       close (50)
c
       return
       end
c
c**********************************************************
